from io import BytesIO, StringIO

import pytest

from vcr import request
from vcr.util import read_body


@pytest.mark.parametrize(
    "input_, expected_output",
    [
        (BytesIO(b"Stream"), b"Stream"),
        (StringIO("Stream"), b"Stream"),
        (iter(["StringIter"]), b"StringIter"),
        (iter(["String", "Iter"]), b"StringIter"),
        (iter([b"BytesIter"]), b"BytesIter"),
        (iter([b"Bytes", b"Iter"]), b"BytesIter"),
        (iter([70, 111, 111]), b"Foo"),
        (iter([]), b""),
        ("String", b"String"),
        (b"Bytes", b"Bytes"),
    ],
)
def test_read_body(input_, expected_output):
    r = request.Request("POST", "http://host.com/", input_, {})
    assert read_body(r) == expected_output


def test_unsupported_read_body():
    r = request.Request("POST", "http://host.com/", iter([[]]), {})
    with pytest.raises(ValueError) as excinfo:
        assert read_body(r)
    assert excinfo.value.args == ("Body type <class 'list'> not supported",)



import contextlib
import http.client as httplib
from io import BytesIO
from tempfile import NamedTemporaryFile
from unittest import mock

from pytest import mark

from vcr import mode, use_cassette
from vcr.cassette import Cassette
from vcr.stubs import VCRHTTPSConnection


class TestVCRConnection:
    def test_setting_of_attributes_get_propagated_to_real_connection(self):
        vcr_connection = VCRHTTPSConnection("www.examplehost.com")
        vcr_connection.ssl_version = "example_ssl_version"
        assert vcr_connection.real_connection.ssl_version == "example_ssl_version"

    @mark.online
    @mock.patch("vcr.cassette.Cassette.can_play_response_for", return_value=False)
    def testing_connect(*args):
        with contextlib.closing(VCRHTTPSConnection("www.google.com")) as vcr_connection:
            vcr_connection.cassette = Cassette("test", record_mode=mode.ALL)
            vcr_connection.real_connection.connect()
            assert vcr_connection.real_connection.sock is not None

    def test_body_consumed_once_stream(self, tmpdir, httpbin):
        self._test_body_consumed_once(
            tmpdir,
            httpbin,
            BytesIO(b"1234567890"),
            BytesIO(b"9876543210"),
            BytesIO(b"9876543210"),
        )

    def test_body_consumed_once_iterator(self, tmpdir, httpbin):
        self._test_body_consumed_once(
            tmpdir,
            httpbin,
            iter([b"1234567890"]),
            iter([b"9876543210"]),
            iter([b"9876543210"]),
        )

    # data2 and data3 should serve the same data, potentially as iterators
    def _test_body_consumed_once(
        self,
        tmpdir,
        httpbin,
        data1,
        data2,
        data3,
    ):
        with NamedTemporaryFile(dir=tmpdir, suffix=".yml") as f:
            testpath = f.name
            # NOTE: ``use_cassette`` is not okay with the file existing
            #       already.  So we using ``.close()`` to not only
            #       close but also delete the empty file, before we start.
            f.close()
            host, port = httpbin.host, httpbin.port
            match_on = ["method", "uri", "body"]
            with use_cassette(testpath, match_on=match_on):
                conn1 = httplib.HTTPConnection(host, port)
                conn1.request("POST", "/anything", body=data1)
                conn1.getresponse()
                conn2 = httplib.HTTPConnection(host, port)
                conn2.request("POST", "/anything", body=data2)
                conn2.getresponse()
            with use_cassette(testpath, match_on=match_on) as cass:
                conn3 = httplib.HTTPConnection(host, port)
                conn3.request("POST", "/anything", body=data3)
                conn3.getresponse()
            assert cass.play_counts[0] == 0
            assert cass.play_counts[1] == 1



"""Test using a proxy."""

import asyncio
import http.server
import socketserver
import threading
from urllib.request import urlopen

import pytest

import vcr

# Conditional imports
requests = pytest.importorskip("requests")


class Proxy(http.server.SimpleHTTPRequestHandler):
    """
    Simple proxy server.

    (Inspired by: http://effbot.org/librarybook/simplehttpserver.htm).
    """

    def do_GET(self):
        upstream_response = urlopen(self.path)
        try:
            status = upstream_response.status
            headers = upstream_response.headers.items()
        except AttributeError:
            # In Python 2 the response is an addinfourl instance.
            status = upstream_response.code
            headers = upstream_response.info().items()
        self.log_request(status)
        self.send_response_only(status, upstream_response.msg)
        for header in headers:
            self.send_header(*header)
        self.end_headers()
        self.copyfile(upstream_response, self.wfile)

    def do_CONNECT(self):
        host, port = self.path.split(":")

        asyncio.run(self._tunnel(host, port, self.connection))

    async def _tunnel(self, host, port, client_sock):
        target_r, target_w = await asyncio.open_connection(host=host, port=port)

        self.send_response(http.HTTPStatus.OK)
        self.end_headers()

        source_r, source_w = await asyncio.open_connection(sock=client_sock)

        async def channel(reader, writer):
            while True:
                data = await reader.read(1024)
                if not data:
                    break
                writer.write(data)
                await writer.drain()

            writer.close()
            await writer.wait_closed()

        await asyncio.gather(
            channel(target_r, source_w),
            channel(source_r, target_w),
        )


@pytest.fixture(scope="session")
def proxy_server():
    with socketserver.ThreadingTCPServer(("", 0), Proxy) as httpd:
        proxy_process = threading.Thread(target=httpd.serve_forever)
        proxy_process.start()
        yield "http://{}:{}".format(*httpd.server_address)
        httpd.shutdown()
        proxy_process.join()


def test_use_proxy(tmpdir, httpbin, proxy_server):
    """Ensure that it works with a proxy."""
    with vcr.use_cassette(str(tmpdir.join("proxy.yaml"))):
        response = requests.get(httpbin.url, proxies={"http": proxy_server})

    with vcr.use_cassette(str(tmpdir.join("proxy.yaml")), mode="none") as cassette:
        cassette_response = requests.get(httpbin.url, proxies={"http": proxy_server})

    assert cassette_response.headers == response.headers
    assert cassette.play_count == 1


def test_use_https_proxy(tmpdir, httpbin_secure, proxy_server):
    """Ensure that it works with an HTTPS proxy."""
    with vcr.use_cassette(str(tmpdir.join("proxy.yaml"))):
        response = requests.get(httpbin_secure.url, proxies={"https": proxy_server})

    with vcr.use_cassette(str(tmpdir.join("proxy.yaml")), mode="none") as cassette:
        cassette_response = requests.get(
            httpbin_secure.url,
            proxies={"https": proxy_server},
        )

    assert cassette_response.headers == response.headers
    assert cassette.play_count == 1

    # The cassette URL points to httpbin, not the proxy
    assert cassette.requests[0].url == httpbin_secure.url + "/"
